#include "Play.hpp"
#include <Graphics/Graphics.hpp>
#include <Graphics/Font.hpp>
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>
#include <memory.h>
#include <netdb.h>
#include <cstdlib>
#include <iostream>
#include <sstream>
#include <string>

// Based on example code from:
// http://people.cs.uchicago.edu/~mark/51081/labs/LAB6/sock.html
// (which appears to no longer exist... oh well.)
int connect_to_server(char *host, char *port) {
	int sock = -1;
	struct addrinfo *serv_addrinfo = NULL;
	cout << "Looking up " << host << ":" << port << "..." << endl;
	int ret = getaddrinfo(host, port, NULL, &serv_addrinfo);
	if (ret != 0) {
		cout << "Host lookup error: " << gai_strerror(ret) << endl;
		return -1;
	}
	for (struct addrinfo *current = serv_addrinfo; current != NULL; 
	    current = current->ai_next) {
		if (current->ai_socktype != SOCK_STREAM) continue; //no UDP, please.
		cout << "Attempting Connection..."; cout.flush();
		sock = socket(current->ai_family, current->ai_socktype, 
		              current->ai_protocol);
		if (sock == -1) {
			cout << " (fails on socket)" << endl;
			continue;
		}
		int res = connect(sock, current->ai_addr, current->ai_addrlen);
		if (res != 0) {
			cout << " (fails on connect)" << endl;
			close(sock);
			sock = -1;
			continue;
		}
		cout << " (success)" << endl;
		break; //we
	}
	freeaddrinfo(serv_addrinfo);

	if (sock == -1) {
		cout << "No connection worked." << endl;
		return -1;
	}

	return sock;
}

void Play::reset_board() {
    for (int i = 0; i < BOARD_WIDTH; i++) {
        for (int j = 0; j < BOARD_HEIGHT; j++) {
            revealed[i][j] = HIDDEN;
        }
    }
    game_over = false;
}

Play::Play(char *hostname, char *port, char *game_name) {
    revealing_x = -1;
    revealing_y = -1;
    reset_board();
    
    // Establish a connection
	sock = connect_to_server(hostname, port);
	if (sock == -1) {
		cout << "Connect error, so quitting" << endl;
	    quit_screen(this);
	    return;
	}

    // Send a handshake
    string message = game_name;
    message = (char)message.size() + message;
    message = 'H' + message;
    const char *msg = message.c_str();
    write_buffer.insert(write_buffer.end(), msg, msg + message.size());
}

Play::~Play() {
    if (sock != -1) {
        close(sock);
    }
}

void draw_quad(int x, int y) {
    glBegin(GL_QUADS);
    glVertex2f(x*BLOCK_WIDTH + BLOCK_BORDER,
               y*BLOCK_HEIGHT + BLOCK_BORDER);
    glVertex2f((x+1)*BLOCK_WIDTH - BLOCK_BORDER,
               y*BLOCK_HEIGHT + BLOCK_BORDER);
    glVertex2f((x+1)*BLOCK_WIDTH - BLOCK_BORDER, 
               (y+1)*BLOCK_HEIGHT - BLOCK_BORDER);
    glVertex2f(x*BLOCK_WIDTH + BLOCK_BORDER,
               (y+1)*BLOCK_HEIGHT - BLOCK_BORDER);
    glEnd();
}

void Play::draw() {
    glClearColor(0.5f, 0.5f, 0.5f, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT);
    glLoadIdentity();

    Graphics::FontRef font = Graphics::get_font("gentium.txf");
    assert(font.ref);

    for (int i = 0; i < BOARD_WIDTH; i++) {
        for (int j = 0; j < BOARD_HEIGHT; j++) {
            if (revealed[i][j] == HIDDEN
                && i == revealing_x && j == revealing_y) {
                // User is holding down mouse over square
                glColor3f(0.6f, 0.6f, 0.6f);
                draw_quad(i,j);
            } else if (revealed[i][j] == HIDDEN) {
                // Square is not revealed
                glColor3f(0.8f, 0.8f, 0.8f);
                draw_quad(i,j);
            } else if (revealed[i][j] == MINE) {
                // Square is a revealed mine
                glColor3f(1.0f, 0.0f, 0.0f);
                draw_quad(i,j);
            } else {
                // Square is a revealed regular tile
                glColor3f(0.3f, 0.3f, 0.3f);
                draw_quad(i,j);
                
                // Only draw non-zero labels
                if (revealed[i][j] != 0) {
                    glColor3f(0.0f, 0.0f, 0.0f); // TODO: make colorful?
                    stringstream block_str;
                    block_str << revealed[i][j]; 
                    font->draw_center(block_str.str(), 
                    make_vector((i+0.5f)*BLOCK_WIDTH, j*BLOCK_HEIGHT),
                    MIN(BLOCK_WIDTH, BLOCK_HEIGHT));
                }
            }
        }
    }

    if (game_over) {
        glColor3f(1.0f, 0.0f, 0.0f);
        font->draw_center("Game over! Someone press 'r' to restart!",
            make_vector(Graphics::screen.x / 2.0f, 
                        Graphics::screen.y / 2.0f),
            50.0f);
    }
}

void Play::update(float elapsed_time) {
    // Write if possible
    if (!write_buffer.empty()) {
        int ret = send(sock, &(write_buffer[0]),
                        write_buffer.size(), 0);
        if (ret == -1) {
            printf("Couldn't send");
            perror("Doing send");
            quit_screen(this);
            return;
        }
        if (ret > 0) {
            assert(ret <= (signed)write_buffer.size());
            write_buffer.erase(write_buffer.begin(),
                                write_buffer.begin() + ret);
        }
    }

    // Read if possible
    const unsigned int BufferSize = 2000;
    static uint8_t buffer[BufferSize];
    int ret = recv(sock, buffer, BufferSize, MSG_DONTWAIT);
    if (ret == -1) {
        if (errno != EAGAIN) {
            perror("Doing recv");
            quit_screen(this);
        }
        return;
    }
    read_buffer.insert(read_buffer.end(), buffer, buffer + ret);

    // Now that we have data, we should do something with it:
    while (read_buffer.size() > 0) {
        uint8_t type = read_buffer[0];
        uint32_t size = read_buffer[1];
        if (read_buffer.size() >= 2 + size) {
            if (type == 'T') {
                string message((char *)&(read_buffer[2]), size);
                cout << message << endl;
            } else if (type == 'R') {
                string msg((char *)&(read_buffer[2]), size);
                int x, y;
                size_t space = msg.find(" ");
                size_t space2 = msg.find(" ", space + 1);
                istringstream x_str_stream(msg.substr(0, space));
                x_str_stream >> x;
                istringstream y_str_stream(msg.substr(space + 1, space2));
                y_str_stream >> y;
                istringstream value_str_stream(
                    msg.substr(space2 + 1, msg.size()-1 - space2));
                value_str_stream >> revealed[x][y];
            } else if (type == 'G') { // Game over
                game_over = true;
            } else if (type == 'S') { // Restart
                reset_board();
                write_buffer.clear();
            } else {
                cout << "Unknown message type '" << (char)type 
                    << "' == " << (int)type << "." << endl;
            }
            
            // Clear message out of buffer
            read_buffer.erase(read_buffer.begin(),
                              read_buffer.begin() + 2 + size);
        }
    }
}

void Play::send_message(string message, char type) {
    message = (char)message.size() + message;
    message = type + message;
    const char *msg = message.c_str();
    write_buffer.insert(write_buffer.end(), msg, msg + message.size());
}

// Ask the server to turn over the tile at (x,y)
void Play::reveal_squares(int x, int y) {
    ostringstream x_int_stream, y_int_stream;
    x_int_stream << x;
    y_int_stream << y;
    send_message(x_int_stream.str() + ' ' + y_int_stream.str(), 'R');
}

void Play::handle_event(SDL_Event const &event) {
	if (event.type == SDL_KEYDOWN) {
		if (event.key.keysym.sym == SDLK_ESCAPE) {
			quit_screen(this);
		} else if (event.key.keysym.sym == SDLK_r) { // Restart
            send_message("",'S');
        }
	} else if (event.type == SDL_MOUSEBUTTONDOWN &&
               event.button.button == SDL_BUTTON_LEFT && !game_over) {
        revealing_x = (BOARD_WIDTH * event.button.x) / Graphics::screen.x;
        revealing_y = BOARD_HEIGHT - 1 -
            (BOARD_HEIGHT * event.button.y) / Graphics::screen.y;
    } else if (event.type == SDL_MOUSEBUTTONUP &&
               event.button.button == SDL_BUTTON_LEFT) {
        int x = (BOARD_WIDTH * event.button.x) / Graphics::screen.x;
        int y = BOARD_HEIGHT - 1 -
            (BOARD_HEIGHT * event.button.y) / Graphics::screen.y;

        // If the user didn't drag away from what they originally clicked,
        // reveal it when they let go
        if (revealing_x == x && revealing_y == y && !game_over) {
            reveal_squares(x, y);
        }

        // Release whatever they clicked
        revealing_x = -1;
        revealing_y = -1;
    }
}


