#include "Client.hpp"
#include <Vector/Vector.hpp>
#include <sys/socket.h>
#include <memory.h>
#include <cstdlib>
#include <map>
#include <sstream>

map< string, Game > games;

// Increments non-mine blocks surrounding the given coordinates
void Game::increment_surrounding(int x, int y) {
    for (int i = -1; i < 2; i++) {
        for (int j = -1; j < 2; j++) {
            if (x + i < 0 || y + j < 0 || 
                x + i >= BOARD_WIDTH || y + j >= BOARD_HEIGHT) {
                continue;
            }
            if (board[x + i][y + j] != MINE) {
                board[x + i][y + j]++;
            }
        }
    }
}

// Create a board with mines and accurate counts in each block
void Game::generate_board() {
    // Initialize board to zeroes
    memset(board, 0, sizeof(int) * BOARD_WIDTH * BOARD_HEIGHT);

    // Initialize revealed to zeroes
    memset(revealed, 0, sizeof(int) * BOARD_WIDTH * BOARD_HEIGHT);

    // Fill with mines
    for (int i = 0; i < NUM_MINES; i++) {
        int x = rand_max(BOARD_WIDTH);
        int y = rand_max(BOARD_HEIGHT);
        if (board[x][y] != MINE) {
            board[x][y] = MINE;
            increment_surrounding(x, y);
        } else {
            i--; // try again
        }
    }
}

void client_read_cb(struct ev_loop *loop, ev_io *watcher, int revents) {
	Client *c = (Client *)watcher->data;
	assert(c);
	const unsigned int BufferSize = 2000;
	static uint8_t buffer[BufferSize];
	int ret = recv(watcher->fd, buffer, BufferSize, 0);
	if (ret == -1) {
		cerr << "[" << c->sock << "] Recv error from client." << endl;
		delete c;
		return;
	}
	if (ret == 0) {
		cerr << "[" << c->sock << "] Client is gone." << endl;
		delete c;
		return;
	}
	assert((unsigned)ret <= BufferSize);

	// Process newly read data
	c->read_buffer.insert(c->read_buffer.end(), buffer, buffer + ret);
	while (c->process_read()) {
		// Keep processing reads until no data left
	}
}

void client_write_cb(struct ev_loop *loop, ev_io *watcher, int revents) {
	Client *c = (Client *)watcher->data;
	assert(c);

	// If we don't have anything to write, stop firing this callback
	if (c->write_buffer.empty()) {
		ev_io_stop(loop, watcher);
		if (c->suicide) {
			delete c;
		}
		return;
	}

	int ret = send(watcher->fd, &(c->write_buffer[0]), 
                   c->write_buffer.size(), 0);
	if (ret == -1) {
		cerr << "[" << c->sock << "] Send error to client." << endl;
		delete c;
		return;
	}
	if (ret > 0) {
		assert(ret <= (signed)c->write_buffer.size());
		c->write_buffer.erase(c->write_buffer.begin(), 
                              c->write_buffer.begin() + ret);
	}
}

Client::Client(int _sock) : sock(_sock), suicide(false), name("") {
	ev_io_init(&read_watcher, client_read_cb, sock, EV_READ);
	ev_io_init(&write_watcher, client_write_cb, sock, EV_WRITE);
	read_watcher.data = this;
	write_watcher.data = this;
	ev_io_start(ev_default_loop(0), &read_watcher);
	ev_io_start(ev_default_loop(0), &write_watcher);
}

Client::~Client() {
	cout << "[" << sock << "] Disconnecting." << endl;

	ev_io_stop(ev_default_loop(0), &read_watcher);
	ev_io_stop(ev_default_loop(0), &write_watcher);

    // Remove ourselves from the game list
    games[name].players.remove(this);

    // If this game is now empty, erase it
    if (games[name].players.empty()) {
        games.erase(name);
    }
}

void Client::send_message(string message, char type) {
	cout << "[" << sock << "] sending message '" << message << "', "
         << type << endl;
	message = (char)message.size() + message;
	message = type + message;
	write_data(message.c_str(), message.size());
}

void Client::send_square(Game g, int x, int y) {
    ostringstream xstr, ystr, val;
    xstr << x;
    ystr << y;
    val << g.board[x][y];
    send_message(xstr.str() + " " + ystr.str() + " " + val.str(),'R');
}

void Client::send_error(string message) {
	if (suicide == true) {
		cerr << "[" << sock << "] Double-suicide!" << endl;
	}
	cout << "[" << sock << "] Suicide: '" << message << "'" << endl;
	message = (char)message.size() + message;
	message = 'T' + message;
	write_data(message.c_str(), message.size());
	suicide = true;
}

void Client::write_data(const void *data, unsigned int length) {
	if (sock == -1) {
		cerr << "Trying to write data to a closed socket" << endl;
		return;
	}
	write_buffer.insert(write_buffer.end(), (uint8_t *)data, 
        (uint8_t *)data + length);

	// Start write_watcher if not already started
	ev_io_start(ev_default_loop(0), &write_watcher);
}

void Client::reveal_squares(Game *g, int x, int y) {
    // If out of bounds or already revealed, return
    if (x < 0 || y < 0 || x >= BOARD_WIDTH || y >= BOARD_HEIGHT ||
        g->revealed[x][y] != 0) {
        return;
    }

    g->revealed[x][y] = 1;

    // Send the revealed square to everyone
    for (list< Client * >::iterator it = g->players.begin();
         it != g->players.end(); it++) {
        (*it)->send_square(*g, x, y);
    }

    // If we hit a mine, send the game over message to everyone
    if (g->board[x][y] == MINE) {
        for (list< Client * >::iterator it = g->players.begin();
             it != g->players.end(); it++) {
            (*it)->send_message("",'G');
            g->game_over = true;
        }
    }

    // If there are no adjacent mines, continue revealing
    if (g->board[x][y] == 0) {
        reveal_squares(g, x-1, y-1);
        reveal_squares(g, x, y-1);
        reveal_squares(g, x+1, y-1);
        reveal_squares(g, x-1, y);
        reveal_squares(g, x+1, y);
        reveal_squares(g, x-1, y+1);
        reveal_squares(g, x, y+1);
        reveal_squares(g, x+1, y+1);
    }
}

bool Client::process_read() {
	// Do we have enough for a header?
	if (read_buffer.size() < 2) {
        return false;
    }

	uint8_t type = read_buffer[0];
	uint32_t size = read_buffer[1];

	// Do we have this whole message?
	if (size + 2U > read_buffer.size()) {
        return false;
    }

	if (name == "") {
		if (type == 'H') { // Get name in handshake
			if (size > 0) {
				name = string((char *)&(read_buffer[2]),
                    (char *)&(read_buffer[2]) + size);
				cout << "[" << sock << "] Trying: '" << name << "'" << endl;

                // Register the player
                games[name].players.push_back(this);

                send_message("Successfully connected!",'T');

                // Is this a new game name someone just created?
                if (games[name].new_game) {
                    games[name].new_game = false;
                    games[name].generate_board();
                } else {
                    // Update the new player on the existing board
                    for (int i = 0; i < BOARD_WIDTH; i++) {
                        for (int j = 0; j < BOARD_HEIGHT; j++) {
                            if (games[name].revealed[i][j]) {
                                send_square(games[name], i, j);
                            }
                        }
                    }

                    // If the game is over, let the new player know
                    if (games[name].game_over) {
                        send_message("",'G');
                    }
                }
            } else {
                send_error("Zero-length handshake");
                return false;
            }
		} else {
			send_error("Expecting handshake");
			return false;
		}
	} else if (type == 'R') { // Reveal a square
        if (!games[name].game_over) {
            vector<string> tokens;
            string msg((char *)&(read_buffer[2]), size);
            size_t space = msg.find(" ");
            if (space == string::npos) {
                send_error("Could not tokenize coordinates");
                return false;
            }
            int x, y;
            istringstream x_str_stream(msg.substr(0, space));
            if (!(x_str_stream >> x)) {
                send_error("First coordinate not a number");
                return false;
            }
            istringstream y_str_stream(
                msg.substr(space + 1, msg.size()-1 - space));
            if (!(y_str_stream >> y)) {
                send_error("Second coordinate not a number");
                return false;
            }
            if (x < 0 || y < 0 || x >= BOARD_WIDTH || y >= BOARD_WIDTH) {
                send_error("Coordinates out of bounds");
                return false;
            }
            reveal_squares(&games[name], x, y);
        }
	} else if (type == 'S') { // Restart board
        if (games[name].game_over) {
            printf("Restarting...\n");
            games[name].game_over = false;
            games[name].generate_board();

            // Ask the clients to restart
            for (list< Client * >::iterator it = 
                 games[name].players.begin();
                 it != games[name].players.end(); it++) {
                (*it)->send_message("",'S');
            }
        }
    } else {
	    send_error("Unknown message type");
        cout << "Unknown message type: " << type << endl;
	    return false;
    }
	
    // Erase message
	read_buffer.erase(read_buffer.begin(), read_buffer.begin() + 2 + size);

	return true;
}
